package spurs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Module Spurs.UtilsSource
Check that arr is sorted from indexes s, e (exclusive)
Sort an array from start, stop (exclusive)
sort keys and vals in place from start to stop, using keys to sort both
sort_like keys vals sorts keys and vals in place, using keys to sort both
swaps arr.i and arr.j in place. Raises if either is an invalid index
Same as swap, but for arrays
Turns arr into the cumulative sum in place
Returns the transpose of 2D Dynarray
Returns the transpose of a 2D Array
Returns the identity of nxn
Returns the range start, stop (exclusive)
binary_search arr x searches for x in arr. If it does not exist, returns None